/* ESP32 Super Simple Streaming Source
  August 2024 Gord Payne
  Based on ESP32 Clock Radio project
  Uses VS-1053 MP3 Decoder module.

  Final circuit includes an audio isolation transformer to reduce ground
  loop interference. May or may not be needed in your application.
  Includes WiFiManager for login credential provisioning and importing
  station IDs and URLs.

  Permits use of characters that can be represented by a 7 segment
  LED character.

  Stores Station IDs and URLs in onboard SPIFFs.

  Press button on back during boot to enter Configuration Mode

  Press station selector knob to save the current station as default/favourite.

  For radio station URLs start with:
  http://fmstream.org/index.php?s=94.7%20country

  NOTE: VS1053_ext library doesn't seem to handle low bitrate streams
  ie. 48Kbits. Station may not play.
  Some https:// stations will not play as it's not through an authorize
  app*/
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"

#include "SPI.h"
#include <vs1053_ext.h>

// SPI and VS 1053 pin definitions
#define SPI_CLK_PIN 18
#define SPI_MISO_PIN 4
#define SPI_MOSI_PIN 16

#define VS1053_CS 19
#define VS1053_DCS 21
#define VS1053_DREQ 5

// instantiate mp3 decoder
VS1053 player(VS1053_CS, VS1053_DCS, VS1053_DREQ, VSPI, SPI_MOSI_PIN, SPI_MISO_PIN, SPI_CLK_PIN);

// Rotary Encoder
#include "AiEsp32RotaryEncoder.h"
#define ROTARY_ENCODER_A_PIN 36 //CLK
#define ROTARY_ENCODER_B_PIN 35 //DT
#define ROTARY_ENCODER_BUTTON_PIN 17 // 33 high on boot. Generates a button click. filter by threshold
#define ROTARY_ENCODER_STEPS 4 // CHANGED FROM 2 FOR homemade encoder module
int lastVal = 0; // previous rotary encoder value;

AiEsp32RotaryEncoder rotaryEncoder = AiEsp32RotaryEncoder(ROTARY_ENCODER_A_PIN, ROTARY_ENCODER_B_PIN, ROTARY_ENCODER_BUTTON_PIN, -1, ROTARY_ENCODER_STEPS);
void IRAM_ATTR readEncoderISR() { // encoder interrupt
  rotaryEncoder.readEncoder_ISR();
}

#define ESP_DRD_USE_SPIFFS true
#include <TM1637Display.h>
// TM1637 Module connection pins (Digital Pins)
#define CLK 26
#define DIO 25
TM1637Display display(CLK, DIO);
//#include <FS.h>
#include <SPIFFS.h>
#include <WiFiManager.h> // for web-based credential and prefs configuration
#include <esp_wifi.h>// native Espressif IDF library
#include "displayText.h" // radio station LED display mapping
#include <ArduinoJson.h> // for settings file structure


// JSON configuration file
#define JSON_CONFIG_FILE "/settings.json"  // settings file
int const jsonFileSize = 1024; // referenced in configSave header class


// configuration settings data
char ssid[20] = ""; // default ssid
char password[20] = ""; // default password
char hsSSID[20] = "hotspot";// not used in this application
char hsPSWD[20] = "hsPassword";// not used in this application
int maxVolume = 21; // max for the amp is 21


#include <EEPROM.h>
// location 2 is default/favourite station index
#define EEPROM_SIZE 3


// Configuration Mode Button
#define cfgBtn 23 // config button
int cfgVal;
long lastPress;
boolean changed = false;// for encoder changed
long lastChangeTime; // for encoder rotation
long cThresh = 1000; // threshold for changing station
long bThresh = 200;
boolean stationDisplay = false;

String cStn;// current station displayed
int ndx; // index for encoder (radio station or hour/minute value)



const int nStations = 10;
// Radio station ID numbers for display
char stations[nStations][5] =
{
  "---",
  "---",
  "---",
  "---",
  "---",
  "---",
  "---",
  "---",
  "---",
  "---"
};

char *new_radio;// current station to connect to
// URLs for the radio stations
char radiopresets[nStations][90] =
{
  "url",
  "url",
  "url",
  "url",
  "url",
  "url",
  "url",
  "url",
  "url"
};
//////////////// END OF SETTINGS DATA

// file containing WiFi Manager code and Save/Load methods
#include "configSave.h"
void setup()
{
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0); //disable brownout detector
  delay(1500); // delay to give amplifier time to start
  Serial.begin (115200);
  pinMode(cfgBtn, INPUT);
  display.setBrightness(0x00);// lowest brightness setting
  display.clear();
  renderText(" on ");
  SPIFFS.begin(true);// check for existence of the config file
  boolean sE = SPIFFS.exists( JSON_CONFIG_FILE);
  SPIFFS.end();
  delay(500);

  rotaryEncoder.begin();
  rotaryEncoder.setup(readEncoderISR);
  rotaryEncoder.setBoundaries(-500, 500, true); //minValue, maxValue, circleValues true|false (when max go to min and vice versa)
  rotaryEncoder.setAcceleration(250);

  // look for configuration request
  cfgVal = digitalRead(cfgBtn);
  if (cfgVal == HIGH) {
    renderText("CFG ");// show in configuration mode
    setupNetworkAndFile();// enter settings configuration mode
  }

  // try to load the existing config file
  boolean fileLoaded;
  if (!sE) { // if missing, create the file
    createConfigFile();
  } else {
    fileLoaded = loadConfigFile();// load the config file
  }
  if (fileLoaded) Serial.println("Preferences loaded");

  EEPROM.begin(EEPROM_SIZE);
  delay(50);

  ndx = EEPROM.read(2);// read saved radio station index
  if (ndx > 9) ndx = 0;

  SPI.setHwCs(true);
  Serial.println("setHwCs done");
  SPI.begin(SPI_CLK_PIN, SPI_MISO_PIN, SPI_MOSI_PIN);  // start SPI before starting decoder
   Serial.println("SPI started");
  delay(1000);
  player.begin();
  delay(1000);
    Serial.println("Decoder Running");
  display.setBrightness(0x00);
  player.setVolume(21);//21 max
  renderText("conn");// indicates attempting to connect to URL

  playAudio();// start preferred station
  cStn = String(stations[ndx]);
  renderText(cStn);// show the current station in the display
  lastChangeTime = millis();
  lastPress = millis();
  
  //  Serial.println("Setup finished.");
}

void loop() {
  cfgVal = digitalRead(cfgBtn);
  if (cfgVal == HIGH && millis() - lastPress > bThresh) {// reboot the board
    renderText("rEbo");
    delay(150);
    ESP.restart();
  }
  player.loop();

  // *** rotary encoder adjustment
  if (rotaryEncoder.encoderChanged()) {
    updateRadio();
  }

  // *** rotary encoder switch pressed - save as favourite station
  if (rotaryEncoder.isEncoderButtonClicked())//&& millis() - lastChangeTime > cThresh)// because 33 high on boot look for enough elapsed time
  {
    EEPROM.write(2, ndx);// store current station index as default/favourite
    EEPROM.commit();
    renderText(" SF ");
    delay(500);
    renderText(cStn);
    lastChangeTime = millis();
  }


  // if enough time has passed since station change,
  //play the current station
  if (changed == true && millis() - lastChangeTime > cThresh) {
    cStn = String(stations[ndx]);
    renderText("conn");
    //offRoutine();// not needed in this applcation
    playAudio();
    renderText(cStn);
    changed = false;
    lastChangeTime = millis();
  }
}


// **** methods


void connectToWiFi() {
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(350);
       Serial.print(".");
  }
    Serial.println("\nWiFi connected");

}

void playAudio() {
  if (WiFi.status() != WL_CONNECTED) {
    connectToWiFi();
  }
  // Serial.println("connection To Radio");
  new_radio = radiopresets[ndx];
  player.connecttohost(new_radio);

}

// called from vs1053 - no need to reference in your own code
void vs1053_showstation(const char *info) {
     Serial.print("SHOWING STATION:      ");
     Serial.println(info);
  stationName(info);
}

// determine if valid station playing
void stationName(String nameIn) {
   Serial.print("stationLength: ");
   Serial.println(nameIn.length());
  if (nameIn.length() < 2) {
    renderText("nonE");
  }
}

void offRoutine(void)// controlled closing of WiFi session if needed
{
  // delay(1000);
  player.stop_mp3client();
  WiFi.disconnect(true);
  WiFi.mode(WIFI_OFF);
}

void updateRadio() { //adjust current radio station
  int theVal = rotaryEncoder.readEncoder();
  if (theVal > lastVal) { // station moving forward
    ndx ++;
    changed = true;
    if (ndx > nStations - 1) {
      ndx = 0;
    }
  } else if (theVal <= lastVal) {
    ndx --;
    changed = true;
    if (ndx < 0) ndx = nStations - 1;
  }
  lastVal = theVal;
  cStn = String(stations[ndx]);
  renderText(cStn);
  lastChangeTime = millis();
}
